module io.stream [Impl] {
   backtrackable(Stream_nb, Stream_b) :-
      Impl:backtrackable(Stream_nb, Stream_b).

   text_position(Stream, File, Position) :-
      Impl:text_position(Stream, File, Position).

   text_position_utf8(Stream, File, Position) :-
      Impl:text_position_utf8(Stream, File, Position).

   list(Stream, List) :-
      Impl:list(Stream, List).

   of_kernel_in(X, Stream) :-
      module Stream {
         read(C) :-
            kernel:io_read(X, C).

         read_term(T) :-
            kernel:io_read_term(X, T).

         read_term(T, Options) :-
            kernel:io_read_term(X, T, Options).

         seek(Method, Count) :-
            kernel:io_seek(X, Method, Count).

         close :-
            kernel:io_close(X).
      },
      data.stream:input_stream(Stream).

   of_kernel_out(X, Stream) :-
      module Stream [X] {
         write(C) :-
            kernel:io_write(X, C).

         flush :-
            kernel:io_flush(X).

         print(Term) :-
            kernel:io_print(X, Term).

         print_endline(Term) :-
            kernel:io_print_endline(X, Term).

         endline :-
            kernel:io_endline(X).

         format(Format, Args) :-
            kernel:io_format(X, Format, Args).

         close :-
            kernel:io_close(X).
      }.
},

module Impl {
   constructor: end_of_file/0, pos/4.

   backtrackable(Stream) = data.list:stream(list(Stream)).

   text_position(Stream, File, Position) :-
      'data.ref.backtrackable':(
         new(0, Offset),
         new(0, Line),
         new(0, BOL)
      ),
      module Position [Stream, File, Offset, Line, BOL] {
         position(pos(File, O, L, B)) :-
            Offset:get(O),
            Line:get(L),
            BOL:get(B).

         read(C) :-
            Stream:read(C),
            data.number.(Offset++),
            (
               C = '\n'
            ->
               Offset:get(O),
               BOL:set(O),
               data.number.(Line++)
            ;
               true
            ).
      }.

   text_position_utf8(Stream, File, Position) :-
      'data.ref.backtrackable':(
         new(0, Offset),
         new(0, Line),
         new(0, BOL)
      ),
      module Position [Stream, File, Offset, Line, BOL] {
         position(pos(File, O, L, B)) :-
            Offset:get(O),
            Line:get(L),
            BOL:get(B).

         read(C) :-
            Stream:read(C),
            (
               C \= 'end_of_file',
               Code = data.char:code(C),
               data.number:(
                  Code < 0x80
               ;
                  Code > 0xC0
               )
            ->
               data.number.(Offset++),
               (
                  C = '\n'
               ->
                  Offset:get(O),
                  BOL:set(O),
                  data.number.(Line++)
               ;
                  true
               )
            ;
               true
            ).
      }.

   list(Stream) =
      data.list:accumulate() {
         get(Char) :-
            Stream:read(Char),
            Char \= end_of_file.
      }.
}
